@@ -21,6 +21,22 @@
  * @short_description: An HTTP request and response.
  * @see_also: #SoupMessageHeaders, #SoupMessageBody
  *
+ * A #SoupMessage represents an HTTP message that is being sent or
+ * received.
+ *
+ * For client-side usage, you would create a #SoupMessage with
+ * soup_message_new() or soup_message_new_from_uri(), set up its
+ * fields appropriate, and send it via a #SoupSession.
+ *
+ * For server-side usage, #SoupServer will create #SoupMessage<!--
+ * -->s automatically for incoming requests, which your application
+ * will receive via handlers.
+ *
+ * Note that libsoup's terminology here does not quite match the HTTP
+ * specification: in RFC 2616, an "HTTP-message" is
+ * <emphasis>either</emphasis> a Request, <emphasis>or</emphasis> a
+ * Response. In libsoup, a #SoupMessage combines both the request and
+ * the response.
  **/
 
 /**
@@ -46,11 +62,16 @@
  * filled in right before libsoup writes the request to the network,
  * but you should not count on this; use soup_message_body_flatten()
  * if you want to ensure that %data is filled in. @response_body's
- * %data will be filled in before #SoupMessage::finished is emitted,
- * unless you set the %SOUP_MESSAGE_OVERWRITE_CHUNKS flag.
+ * %data will be filled in before #SoupMessage::finished is emitted.
  *
  * For a server-side #SoupMessage, @request_body's %data will be
  * filled in before #SoupMessage::got_body is emitted.
+ *
+ * To prevent the %data field from being filled in at all (eg, if you
+ * are handling the data from a #SoupMessage::got_chunk, and so don't
+ * need to see it all at the end), call
+ * soup_message_body_set_accumulate() on @response_body or
+ * @request_body as appropriate, passing %FALSE.
  **/
 
 G_DEFINE_TYPE (SoupMessage, soup_message, G_TYPE_OBJECT)
@@ -59,6 +80,7 @@ enum {
 	WROTE_INFORMATIONAL,
 	WROTE_HEADERS,
 	WROTE_CHUNK,
+	WROTE_BODY_DATA,
 	WROTE_BODY,
 
 	GOT_INFORMATIONAL,
@@ -81,6 +103,7 @@ enum {
 	PROP_URI,
 	PROP_HTTP_VERSION,
 	PROP_FLAGS,
+	PROP_SERVER_SIDE,
 	PROP_STATUS_CODE,
 	PROP_REASON_PHRASE,
 
@@ -102,7 +125,7 @@ soup_message_init (SoupMessage *msg)
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 
 	priv->io_status = SOUP_MESSAGE_IO_STATUS_IDLE;
-	priv->http_version = SOUP_HTTP_1_1;
+	priv->http_version = priv->orig_http_version = SOUP_HTTP_1_1;
 
 	msg->request_body = soup_message_body_new ();
 	msg->request_headers = soup_message_headers_new (SOUP_MESSAGE_HEADERS_REQUEST);
@@ -117,6 +140,8 @@ finalize (GObject *object)
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 
 	soup_message_io_cleanup (msg);
+	if (priv->chunk_allocator_dnotify)
+		priv->chunk_allocator_dnotify (priv->chunk_allocator_data);
 
 	if (priv->uri)
 		soup_uri_free (priv->uri);
@@ -131,7 +156,7 @@ finalize (GObject *object)
 	soup_message_body_free (msg->response_body);
 	soup_message_headers_free (msg->response_headers);
 
-	g_free ((char *) msg->reason_phrase);
+	g_free (msg->reason_phrase);
 
 	G_OBJECT_CLASS (soup_message_parent_class)->finalize (object);
 }
@@ -194,6 +219,13 @@ soup_message_class_init (SoupMessageClass *message_class)
 	 * @msg: the message
 	 *
 	 * Emitted immediately after writing a body chunk for a message.
+	 *
+	 * Note that this signal is not parallel to
+	 * #SoupMessage::got_chunk; it is emitted only when a complete
+	 * chunk (added with soup_message_body_append() or
+	 * soup_message_body_append_buffer()) has been written. To get
+	 * more useful continuous progress information, use
+	 * #SoupMessage::wrote_body_data.
 	 **/
 	signals[WROTE_CHUNK] =
 		g_signal_new ("wrote_chunk",
@@ -204,6 +236,28 @@ soup_message_class_init (SoupMessageClass *message_class)
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
 
+	/**
+	 * SoupMessage::wrote-body-data:
+	 * @msg: the message
+	 * @chunk: the data written
+	 *
+	 * Emitted immediately after writing a portion of the message
+	 * body to the network.
+	 *
+	 * Unlike #SoupMessage::wrote_chunk, this is emitted after
+	 * every successful write() call, not only after finishing a
+	 * complete "chunk".
+	 **/
+	signals[WROTE_BODY_DATA] =
+		g_signal_new ("wrote_body_data",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0, /* FIXME after next ABI break */
+			      NULL, NULL,
+			      soup_marshal_NONE__BOXED,
+			      G_TYPE_NONE, 1,
+			      SOUP_TYPE_BUFFER);
+
 	/**
 	 * SoupMessage::wrote-body:
 	 * @msg: the message
@@ -301,7 +355,13 @@ soup_message_class_init (SoupMessageClass *message_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__BOXED,
 			      G_TYPE_NONE, 1,
-			      SOUP_TYPE_BUFFER);
+			      /* Use %G_SIGNAL_TYPE_STATIC_SCOPE so that
+			       * the %SOUP_MEMORY_TEMPORARY buffers used
+			       * by soup-message-io.c when emitting this
+			       * signal don't get forcibly copied by
+			       * g_signal_emit().
+			       */
+			      SOUP_TYPE_BUFFER | G_SIGNAL_TYPE_STATIC_SCOPE);
 
 	/**
 	 * SoupMessage::got-body:
@@ -391,6 +451,13 @@ soup_message_class_init (SoupMessageClass *message_class)
 				    SOUP_TYPE_MESSAGE_FLAGS,
 				    0,
 				    G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_SERVER_SIDE,
+		g_param_spec_boolean (SOUP_MESSAGE_SERVER_SIDE,
+				      "Server-side",
+				      "Whether or not the message is server-side rather than client-side",
+				      FALSE,
+				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (
 		object_class, PROP_STATUS_CODE,
 		g_param_spec_uint (SOUP_MESSAGE_STATUS_CODE,
@@ -412,6 +479,7 @@ set_property (GObject *object, guint prop_id,
 	      const GValue *value, GParamSpec *pspec)
 {
 	SoupMessage *msg = SOUP_MESSAGE (object);
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 
 	switch (prop_id) {
 	case PROP_METHOD:
@@ -426,6 +494,9 @@ set_property (GObject *object, guint prop_id,
 	case PROP_FLAGS:
 		soup_message_set_flags (msg, g_value_get_flags (value));
 		break;
+	case PROP_SERVER_SIDE:
+		priv->server_side = g_value_get_boolean (value);
+		break;
 	case PROP_STATUS_CODE:
 		soup_message_set_status (msg, g_value_get_uint (value));
 		break;
@@ -434,6 +505,7 @@ set_property (GObject *object, guint prop_id,
 					      g_value_get_string (value));
 		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -458,6 +530,9 @@ get_property (GObject *object, guint prop_id,
 	case PROP_FLAGS:
 		g_value_set_flags (value, priv->msg_flags);
 		break;
+	case PROP_SERVER_SIDE:
+		g_value_set_boolean (value, priv->server_side);
+		break;
 	case PROP_STATUS_CODE:
 		g_value_set_uint (value, msg->status_code);
 		break;
@@ -465,6 +540,7 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_string (value, msg->reason_phrase);
 		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -625,6 +701,20 @@ soup_message_wrote_chunk (SoupMessage *msg)
 	g_signal_emit (msg, signals[WROTE_CHUNK], 0);
 }
 
+/**
+ * soup_message_wrote_body_data:
+ * @msg: a #SoupMessage
+ * @chunk: the data written
+ *
+ * Emits the %wrote_body_data signal, indicating that the IO layer
+ * finished writing a portion of @msg's body.
+ **/
+void
+soup_message_wrote_body_data (SoupMessage *msg, SoupBuffer *chunk)
+{
+	g_signal_emit (msg, signals[WROTE_BODY_DATA], 0, chunk);
+}
+
 /**
  * soup_message_wrote_body:
  * @msg: a #SoupMessage
@@ -682,15 +772,13 @@ static void
 got_body (SoupMessage *req)
 {
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (req);
+	SoupMessageBody *body;
 
-	if (!(priv->msg_flags & SOUP_MESSAGE_OVERWRITE_CHUNKS)) {
+	body = priv->server_side ? req->request_body : req->response_body;
+	if (soup_message_body_get_accumulate (body)) {
 		SoupBuffer *buffer;
 
-		/* Figure out *which* body we read, and flatten it. */
-		if (req->status_code == 0)
-			buffer = soup_message_body_flatten (req->request_body);
-		else
-			buffer = soup_message_body_flatten (req->response_body);
+		buffer = soup_message_body_flatten (body);
 		soup_buffer_free (buffer);
 	}
 }
@@ -768,13 +856,7 @@ header_handler_metamarshal (GClosure *closure, GValue *return_value,
 	if (priv->io_status != SOUP_MESSAGE_IO_STATUS_RUNNING)
 		return;
 
-	/* If status_code is SOUP_STATUS_NONE, we're still processing
-	 * the request side; if it's not, we're processing the
-	 * response side.
-	 */
-	hdrs = (msg->status_code == SOUP_STATUS_NONE) ?
-		msg->request_headers : msg->response_headers;
-
+	hdrs = priv->server_side ? msg->request_headers : msg->response_headers;
 	if (soup_message_headers_get (hdrs, header_name)) {
 		closure->marshal (closure, return_value, n_param_values,
 				  param_values, invocation_hint,
@@ -1012,27 +1094,30 @@ soup_message_get_proxy_auth (SoupMessage *msg)
 void
 soup_message_cleanup_response (SoupMessage *req)
 {
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (req);
+
 	soup_message_body_truncate (req->response_body);
 	soup_message_headers_clear (req->response_headers);
 
 	req->status_code = SOUP_STATUS_NONE;
 	if (req->reason_phrase) {
-		g_free ((char *) req->reason_phrase);
+		g_free (req->reason_phrase);
 		req->reason_phrase = NULL;
 	}
+	priv->http_version = priv->orig_http_version;
+
 	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_STATUS_CODE);
 	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_REASON_PHRASE);
+	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_HTTP_VERSION);
 }
 
 /**
  * SoupMessageFlags:
  * @SOUP_MESSAGE_NO_REDIRECT: The session should not follow redirect
  * (3xx) responses received by this message.
- * @SOUP_MESSAGE_OVERWRITE_CHUNKS: Each chunk of the response will be
- * freed after its corresponding %got_chunk signal is emitted, meaning
- * %response will still be empty after the message is complete. You
- * can use this to save memory if you expect the response to be large
- * and you are able to process it a chunk at a time.
+ * @SOUP_MESSAGE_OVERWRITE_CHUNKS: Deprecated: equivalent to calling
+ * soup_message_body_set_accumulate() on the incoming message body
+ * (ie, %response_body for a client-side request), passing %FALSE.
  *
  * Various flags that can be set on a #SoupMessage to alter its
  * behavior.
@@ -1048,9 +1133,18 @@ soup_message_cleanup_response (SoupMessage *req)
 void
 soup_message_set_flags (SoupMessage *msg, SoupMessageFlags flags)
 {
-	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	SoupMessagePrivate *priv;
 
-	SOUP_MESSAGE_GET_PRIVATE (msg)->msg_flags = flags;
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	if ((priv->msg_flags ^ flags) & SOUP_MESSAGE_OVERWRITE_CHUNKS) {
+		soup_message_body_set_accumulate (
+			priv->server_side ? msg->request_body : msg->response_body,
+			!(flags & SOUP_MESSAGE_OVERWRITE_CHUNKS));
+	}
+
+	priv->msg_flags = flags;
 	g_object_notify (G_OBJECT (msg), SOUP_MESSAGE_FLAGS);
 }
 
@@ -1090,9 +1184,14 @@ soup_message_get_flags (SoupMessage *msg)
 void
 soup_message_set_http_version (SoupMessage *msg, SoupHTTPVersion version)
 {
-	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	SoupMessagePrivate *priv;
 
-	SOUP_MESSAGE_GET_PRIVATE (msg)->http_version = version;
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	priv->http_version = version;
+	if (msg->status_code == SOUP_STATUS_NONE)
+		priv->orig_http_version = version;
 	g_object_notify (G_OBJECT (msg), SOUP_MESSAGE_HTTP_VERSION);
 }
 
@@ -1216,7 +1315,7 @@ soup_message_set_status (SoupMessage *msg, guint status_code)
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 	g_return_if_fail (status_code != 0);
 
-	g_free ((char *) msg->reason_phrase);
+	g_free (msg->reason_phrase);
 
 	msg->status_code = status_code;
 	msg->reason_phrase = g_strdup (soup_status_get_phrase (status_code));
@@ -1241,7 +1340,7 @@ soup_message_set_status_full (SoupMessage *msg,
 	g_return_if_fail (status_code != 0);
 	g_return_if_fail (reason_phrase != NULL);
 
-	g_free ((char *) msg->reason_phrase);
+	g_free (msg->reason_phrase);
 
 	msg->status_code = status_code;
 	msg->reason_phrase = g_strdup (reason_phrase);
@@ -1265,3 +1364,81 @@ soup_message_get_io_status (SoupMessage *msg)
 
 	return priv->io_status;
 }
+
+/**
+ * SoupChunkAllocator:
+ * @msg: the #SoupMessage the chunk is being allocated for
+ * @max_len: the maximum length that will be read, or 0.
+ * @user_data: the data passed to soup_message_set_chunk_allocator()
+ *
+ * The prototype for a chunk allocation callback. This should allocate
+ * a new #SoupBuffer and return it for the I/O layer to read message
+ * body data off the network into.
+ *
+ * If @max_len is non-0, it indicates the maximum number of bytes that
+ * could be read, based on what is known about the message size. Note
+ * that this might be a very large number, and you should not simply
+ * try to allocate that many bytes blindly. If @max_len is 0, that
+ * means that libsoup does not know how many bytes remain to be read,
+ * and the allocator should return a buffer of a size that it finds
+ * convenient.
+ *
+ * If the allocator returns %NULL, the message will be paused. It is
+ * up to the application to make sure that it gets unpaused when it
+ * becomes possible to allocate a new buffer.
+ *
+ * Return value: the new buffer (or %NULL)
+ **/
+
+/**
+ * soup_message_set_chunk_allocator:
+ * @msg: a #SoupMessage
+ * @allocator: the chunk allocator callback
+ * @user_data: data to pass to @allocator
+ * @destroy_notify: destroy notifier to free @user_data when @msg is
+ * destroyed
+ *
+ * Sets an alternate chunk-allocation function to use when reading
+ * @msg's body. Every time data is available to read, libsoup will
+ * call @allocator, which should return a #SoupBuffer. (See
+ * #SoupChunkAllocator for additional details.) Libsoup will then read
+ * data from the network into that buffer, and update the buffer's
+ * %length to indicate how much data it read.
+ *
+ * Generally, a custom chunk allocator would be used in conjunction
+ * with soup_message_body_set_accumulate() %FALSE and
+ * #SoupMessage::got_chunk, as part of a strategy to avoid unnecessary
+ * copying of data. However, you cannot assume that every call to the
+ * allocator will be followed by a call to your %got_chunk handler; if
+ * an I/O error occurs, then the buffer will be unreffed without ever
+ * having been used. If your buffer-allocation strategy requires
+ * special cleanup, use soup_buffer_new_with_owner() rather than doing
+ * the cleanup from the %got_chunk handler.
+ *
+ * The other thing to remember when using non-accumulating message
+ * bodies is that the buffer passed to the %got_chunk handler will be
+ * unreffed after the handler returns, just as it would be in the
+ * non-custom-allocated case. If you want to hand the chunk data off
+ * to some other part of your program to use later, you'll need to ref
+ * the #SoupBuffer (or its owner, in the soup_buffer_new_with_owner()
+ * case) to ensure that the data remains valid.
+ **/
+void
+soup_message_set_chunk_allocator (SoupMessage *msg,
+				  SoupChunkAllocator allocator,
+				  gpointer user_data,
+				  GDestroyNotify destroy_notify)
+{
+	SoupMessagePrivate *priv;
+
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
+	priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	if (priv->chunk_allocator_dnotify)
+		priv->chunk_allocator_dnotify (priv->chunk_allocator_data);
+
+	priv->chunk_allocator         = allocator;
+	priv->chunk_allocator_data    = user_data;
+	priv->chunk_allocator_dnotify = destroy_notify;
+}
