@@ -68,8 +68,6 @@ typedef struct {
 	SoupURI *proxy_uri;
 	SoupAuth *proxy_auth;
 
-	guint max_conns, max_conns_per_host;
-
 	char *ssl_ca_file;
 	SoupSSLCredentials *ssl_creds;
 
@@ -83,6 +81,8 @@ typedef struct {
 	GHashTable *hosts; /* SoupURI -> SoupSessionHost */
 	GHashTable *conns; /* SoupConnection -> SoupSessionHost */
 	guint num_conns;
+	guint max_conns, max_conns_per_host;
+	guint io_timeout, idle_timeout;
 
 	/* Must hold the host_lock before potentially creating a
 	 * new SoupSessionHost, or adding/removing a connection.
@@ -91,11 +91,6 @@ typedef struct {
 	GMutex *host_lock;
 
 	GMainContext *async_context;
-
-	/* Holds the timeout value for the connection, when
-	   no response is received.
-	*/
-	guint timeout;
 } SoupSessionPrivate;
 #define SOUP_SESSION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SESSION, SoupSessionPrivate))
 
@@ -121,7 +116,9 @@ extern SoupURI  *soup_uri_copy_root  (SoupURI *uri);
 G_DEFINE_TYPE (SoupSession, soup_session, G_TYPE_OBJECT)
 
 enum {
+	REQUEST_QUEUED,
 	REQUEST_STARTED,
+	REQUEST_UNQUEUED,
 	AUTHENTICATE,
 	LAST_SIGNAL
 };
@@ -139,6 +136,7 @@ enum {
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_USER_AGENT,
+	PROP_IDLE_TIMEOUT,
 
 	LAST_PROP
 };
@@ -163,8 +161,6 @@ soup_session_init (SoupSession *session)
 	priv->max_conns = SOUP_SESSION_MAX_CONNS_DEFAULT;
 	priv->max_conns_per_host = SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT;
 
-	priv->timeout = 0;
-
 	priv->auth_manager = soup_auth_manager_new (session);
 	soup_auth_manager_add_type (priv->auth_manager, SOUP_TYPE_AUTH_BASIC);
 	soup_auth_manager_add_type (priv->auth_manager, SOUP_TYPE_AUTH_DIGEST);
@@ -216,6 +212,8 @@ finalize (GObject *object)
 	g_hash_table_destroy (priv->hosts);
 	g_hash_table_destroy (priv->conns);
 
+	g_free (priv->user_agent);
+
 	soup_auth_manager_free (priv->auth_manager);
 	if (priv->ntlm_manager)
 		soup_auth_manager_ntlm_free (priv->ntlm_manager);
@@ -252,13 +250,66 @@ soup_session_class_init (SoupSessionClass *session_class)
 
 	/* signals */
 
+	/**
+	 * SoupSession::request-queued:
+	 * @session: the session
+	 * @msg: the request that was queued
+	 *
+	 * Emitted when a request is queued on @session. (Note that
+	 * "queued" doesn't just mean soup_session_queue_message();
+	 * soup_session_send_message() implicitly queues the message
+	 * as well.)
+	 *
+	 * When sending a request, first #SoupSession::request_queued
+	 * is emitted, indicating that the session has become aware of
+	 * the request.
+	 *
+	 * Once a connection is available to send the request on, the
+	 * session emits #SoupSession::request_started. Then, various
+	 * #SoupMessage signals are emitted as the message is
+	 * processed. If the message is requeued, it will emit
+	 * #SoupMessage::restarted, which will then be followed by
+	 * another #SoupSession::request_started and another set of
+	 * #SoupMessage signals when the message is re-sent.
+	 *
+	 * Eventually, the message will emit #SoupMessage::finished.
+	 * Normally, this signals the completion of message
+	 * processing. However, it is possible that the application
+	 * will requeue the message from the "finished" handler (or
+	 * equivalently, from the soup_session_queue_message()
+	 * callback). In that case, the process will loop back to
+	 * #SoupSession::request_started.
+	 *
+	 * Eventually, a message will reach "finished" and not be
+	 * requeued. At that point, the session will emit
+	 * #SoupSession::request_unqueued to indicate that it is done
+	 * with the message.
+	 *
+	 * To sum up: #SoupSession::request_queued and
+	 * #SoupSession::request_unqueued are guaranteed to be emitted
+	 * exactly once, but #SoupSession::request_started and
+	 * #SoupMessage::finished (and all of the other #SoupMessage
+	 * signals) may be invoked multiple times for a given message.
+	 **/
+	signals[REQUEST_QUEUED] =
+		g_signal_new ("request-queued",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0, /* FIXME? */
+			      NULL, NULL,
+			      soup_marshal_NONE__OBJECT,
+			      G_TYPE_NONE, 1,
+			      SOUP_TYPE_MESSAGE);
+
 	/**
 	 * SoupSession::request-started:
 	 * @session: the session
 	 * @msg: the request being sent
 	 * @socket: the socket the request is being sent on
 	 *
-	 * Emitted just before a request is sent.
+	 * Emitted just before a request is sent. See
+	 * #SoupSession::request_queued for a detailed description of
+	 * the message lifecycle within a session.
 	 **/
 	signals[REQUEST_STARTED] =
 		g_signal_new ("request-started",
@@ -271,6 +322,26 @@ soup_session_class_init (SoupSessionClass *session_class)
 			      SOUP_TYPE_MESSAGE,
 			      SOUP_TYPE_SOCKET);
 
+	/**
+	 * SoupSession::request-unqueued:
+	 * @session: the session
+	 * @msg: the request that was unqueued
+	 *
+	 * Emitted when a request is removed from @session's queue,
+	 * indicating that @session is done with it. See
+	 * #SoupSession::request_queued for a detailed description of the
+	 * message lifecycle within a session.
+	 **/
+	signals[REQUEST_UNQUEUED] =
+		g_signal_new ("request-unqueued",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0, /* FIXME? */
+			      NULL, NULL,
+			      soup_marshal_NONE__OBJECT,
+			      G_TYPE_NONE, 1,
+			      SOUP_TYPE_MESSAGE);
+
 	/**
 	 * SoupSession::authenticate:
 	 * @session: the session
@@ -332,6 +403,13 @@ soup_session_class_init (SoupSessionClass *session_class)
 				  G_MAXINT,
 				  SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT,
 				  G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_IDLE_TIMEOUT,
+		g_param_spec_uint (SOUP_SESSION_IDLE_TIMEOUT,
+				   "Idle Timeout",
+				   "Connection lifetime when idle",
+				   0, G_MAXUINT, 0,
+				   G_PARAM_READWRITE));
 	g_object_class_install_property (
 		object_class, PROP_USE_NTLM,
 		g_param_spec_boolean (SOUP_SESSION_USE_NTLM,
@@ -491,7 +569,7 @@ set_property (GObject *object, guint prop_id,
 			g_main_context_ref (priv->async_context);
 		break;
 	case PROP_TIMEOUT:
-		priv->timeout = g_value_get_uint (value);
+		priv->io_timeout = g_value_get_uint (value);
 		break;
 	case PROP_USER_AGENT:
 		g_free (priv->user_agent);
@@ -508,7 +586,11 @@ set_property (GObject *object, guint prop_id,
 		} else
 			priv->user_agent = g_strdup (user_agent);
 		break;
+	case PROP_IDLE_TIMEOUT:
+		priv->idle_timeout = g_value_get_uint (value);
+		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -540,12 +622,16 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
 	case PROP_TIMEOUT:
-		g_value_set_uint (value, priv->timeout);
+		g_value_set_uint (value, priv->io_timeout);
 		break;
 	case PROP_USER_AGENT:
 		g_value_set_string (value, priv->user_agent);
 		break;
+	case PROP_IDLE_TIMEOUT:
+		g_value_set_uint (value, priv->idle_timeout);
+		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -645,14 +731,15 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 	g_return_if_fail (new_loc != NULL);
 
 	if (msg->status_code == SOUP_STATUS_MOVED_PERMANENTLY ||
-	    msg->status_code == SOUP_STATUS_FOUND ||
 	    msg->status_code == SOUP_STATUS_TEMPORARY_REDIRECT) {
 		/* Don't redirect non-safe methods */
 		if (msg->method != SOUP_METHOD_GET &&
 		    msg->method != SOUP_METHOD_HEAD &&
-		    msg->method != SOUP_METHOD_OPTIONS)
+		    msg->method != SOUP_METHOD_OPTIONS &&
+		    msg->method != SOUP_METHOD_PROPFIND)
 			return;
-	} else if (msg->status_code == SOUP_STATUS_SEE_OTHER) {
+	} else if (msg->status_code == SOUP_STATUS_SEE_OTHER ||
+		   msg->status_code == SOUP_STATUS_FOUND) {
 		/* Redirect using a GET */
 		g_object_set (msg,
 			      SOUP_MESSAGE_METHOD, SOUP_METHOD_GET,
@@ -711,59 +798,51 @@ connection_started_request (SoupConnection *conn, SoupMessage *msg,
 					      "User-Agent", priv->user_agent);
 	}
 
+	/* Kludge to deal with the fact that CONNECT msgs come from the
+	 * SoupConnection rather than being queued normally.
+	 */
+	if (msg->method == SOUP_METHOD_CONNECT)
+		g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
+
 	g_signal_emit (session, signals[REQUEST_STARTED], 0,
 		       msg, soup_connection_get_socket (conn));
 }
 
-static void
-find_oldest_connection (gpointer key, gpointer host, gpointer data)
-{
-	SoupConnection *conn = key, **oldest = data;
-
-	/* Don't prune a connection that is currently in use, or
-	 * hasn't been used yet.
-	 */
-	if (soup_connection_is_in_use (conn) ||
-	    soup_connection_last_used (conn) == 0)
-		return;
-
-	if (!*oldest || (soup_connection_last_used (conn) <
-			 soup_connection_last_used (*oldest)))
-		*oldest = conn;
-}
-
-/**
- * soup_session_try_prune_connection:
- * @session: a #SoupSession
- *
- * Finds the least-recently-used idle connection in @session and closes
- * it.
- *
- * Return value: %TRUE if a connection was closed, %FALSE if there are
- * no idle connections.
- **/
 gboolean
 soup_session_try_prune_connection (SoupSession *session)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
-	SoupConnection *oldest = NULL;
+	GPtrArray *conns;
+	GHashTableIter iter;
+	gpointer conn, host;
+	int i;
+
+	conns = g_ptr_array_new ();
 
 	g_mutex_lock (priv->host_lock);
-	g_hash_table_foreach (priv->conns, find_oldest_connection,
-			      &oldest);
-	if (oldest) {
-		/* Ref the connection before unlocking the mutex in
-		 * case someone else tries to prune it too.
+	g_hash_table_iter_init (&iter, priv->conns);
+	while (g_hash_table_iter_next (&iter, &conn, &host)) {
+		/* Don't prune a connection that is currently in use,
+		 * or hasn't been used yet.
 		 */
-		g_object_ref (oldest);
-		g_mutex_unlock (priv->host_lock);
-		soup_connection_disconnect (oldest);
-		g_object_unref (oldest);
-		return TRUE;
-	} else {
-		g_mutex_unlock (priv->host_lock);
+		if (!soup_connection_is_in_use (conn) &&
+		    soup_connection_last_used (conn) > 0)
+			g_ptr_array_add (conns, g_object_ref (conn));
+	}
+	g_mutex_unlock (priv->host_lock);
+
+	if (!conns->len) {
+		g_ptr_array_free (conns, TRUE);
 		return FALSE;
 	}
+
+	for (i = 0; i < conns->len; i++) {
+		soup_connection_disconnect (conns->pdata[i]);
+		g_object_unref (conns->pdata[i]);
+	}
+	g_ptr_array_free (conns, TRUE);
+
+	return TRUE;
 }
 
 static void
@@ -931,7 +1010,8 @@ soup_session_get_connection (SoupSession *session, SoupMessage *msg,
 		SOUP_CONNECTION_PROXY_URI, priv->proxy_uri,
 		SOUP_CONNECTION_SSL_CREDENTIALS, priv->ssl_creds,
 		SOUP_CONNECTION_ASYNC_CONTEXT, priv->async_context,
-		SOUP_CONNECTION_TIMEOUT, priv->timeout,
+		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
+		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
 		NULL);
 	g_signal_connect (conn, "connect_result",
 			  G_CALLBACK (connect_result),
@@ -979,6 +1059,8 @@ message_finished (SoupMessage *msg, gpointer user_data)
 	if (!SOUP_MESSAGE_IS_STARTING (msg)) {
 		soup_message_queue_remove_message (priv->queue, msg);
 		g_signal_handlers_disconnect_by_func (msg, message_finished, session);
+		g_signal_handlers_disconnect_by_func (msg, redirect_handler, session);
+		g_signal_emit (session, signals[REQUEST_UNQUEUED], 0, msg);
 	}
 }
 
@@ -999,6 +1081,8 @@ queue_message (SoupSession *session, SoupMessage *msg,
 
 	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
 	soup_message_queue_append (priv->queue, msg);
+
+	g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
 }
 
 /**
