@@ -14,15 +14,13 @@
 
 #include "test-utils.h"
 
-GMainLoop *loop;
-
 typedef struct {
 	const char *method;
 	const char *path;
 	guint status_code;
 } TestRequest;
 
-struct {
+static struct {
 	TestRequest requests[3];
 } tests[] = {
 	/* A redirecty response to a GET should cause a redirect */
@@ -177,8 +175,6 @@ do_redirect_tests (SoupURI *base_uri)
 	g_object_unref (session);
 }
 
-GThread *server_thread;
-
 static void
 server_callback (SoupServer *server, SoupMessage *msg,
 		 const char *path, GHashTable *query,
@@ -199,6 +195,14 @@ server_callback (SoupServer *server, SoupMessage *msg,
 			return;
 		}
 
+		/* Make sure that a HTTP/1.0 redirect doesn't cause an
+		 * HTTP/1.0 re-request. (#521848)
+		 */
+		if (soup_message_get_http_version (msg) == SOUP_HTTP_1_0) {
+			soup_message_set_status (msg, SOUP_STATUS_BAD_REQUEST);
+			return;
+		}
+
 		soup_message_set_status (msg, SOUP_STATUS_OK);
 		soup_message_set_response (msg, "text/plain",
 					   SOUP_MEMORY_STATIC,
@@ -213,6 +217,9 @@ server_callback (SoupServer *server, SoupMessage *msg,
 		return;
 	}
 
+	/* See above comment re bug 521848. */
+	soup_message_set_http_version (msg, SOUP_HTTP_1_0);
+
 	soup_message_set_status (msg, status_code);
 	if (*remainder) {
 		soup_message_headers_replace (msg->response_headers,
@@ -223,7 +230,7 @@ server_callback (SoupServer *server, SoupMessage *msg,
 	}
 }
 
-gboolean run_tests = TRUE;
+static gboolean run_tests = TRUE;
 
 static GOptionEntry no_test_entry[] = {
         { "no-tests", 'n', G_OPTION_FLAG_NO_ARG | G_OPTION_FLAG_REVERSE,
